//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KeyboardGUI class - creates a frame to display an animated keyboard
//
import java.awt.Color;
import java.awt.Frame;
import java.awt.Font;
import java.awt.Panel;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.CheckboxMenuItem;
import java.awt.Label;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.MenuShortcut;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.text.MessageFormat;

/**
 * Displays an animated keyboard graphic, using keycodes arranged in the
 * following pattern:<br>
 * <pre>
 * -----------------------Main keyboard---------------------     Numeric keybd
 *  41,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13,   14  |  69, 98, 55, 74
 *   15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27,  43  |  71, 72, 73
 *   58,  30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,     28  |  75, 76, 77, 78
 *    42,  44, 45, 46, 47, 48, 49, 50, 51, 52, 53,        54  |  79, 80, 81
 *   29,    56, ------------57-------------, 100,         97  |      82, 83, 96
 * </pre>
 */
public class KeyboardGUI extends Frame
			implements Runnable,
			ItemListener, ActionListener, WindowListener {
	private boolean inAnApplet = true;

	// message and colour catalog information
	private static final ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );
	private static final ResourceBundle	colours		=
				ResourceBundle.getBundle( "ColorResource" );

	// menubar GUI components
	private MenuBar			menu_bar;
	private Menu			file_menu;
	private MenuItem		file_close;
	private Menu			options_menu;
	private CheckboxMenuItem	options_expected, options_colourize,
					options_legends,
					options_error, options_success;

	// key keycodes and ranges; note - when this module and others talk
	// about keycodes, these are really only a convenient way to assign
	// unique numbers to keyboard positions; although the values used
	// are real keycodes, they could have been any unique scheme
	private static final KeyCode	ROW_1_KEY1	= KeyCode.KC_41;
	private static final KeyCode[]	ROW_1_SEQ = {
			KeyCode.KC_2,  KeyCode.KC_3,  KeyCode.KC_4,
			KeyCode.KC_5,  KeyCode.KC_6,  KeyCode.KC_7,
			KeyCode.KC_8,  KeyCode.KC_9,  KeyCode.KC_10,
			KeyCode.KC_11, KeyCode.KC_12, KeyCode.KC_13 };
	private static final KeyCode	ROW_1_BACKSPACE	= KeyCode.KC_14;
	private static final KeyCode	ROW_2_TAB	= KeyCode.KC_15;
	private static final KeyCode[]	ROW_2_SEQ = {
			KeyCode.KC_16, KeyCode.KC_17, KeyCode.KC_18,
			KeyCode.KC_19, KeyCode.KC_20, KeyCode.KC_21,
			KeyCode.KC_22, KeyCode.KC_23, KeyCode.KC_24,
			KeyCode.KC_25, KeyCode.KC_26, KeyCode.KC_27 };
	private static final KeyCode	ROW_2_KEY14	= KeyCode.KC_43;
	private static final KeyCode	ROW_3_CAPSLOCK	= KeyCode.KC_58;
	private static final KeyCode[]	ROW_3_SEQ = {
			KeyCode.KC_30, KeyCode.KC_31, KeyCode.KC_32,
			KeyCode.KC_33, KeyCode.KC_34, KeyCode.KC_35,
			KeyCode.KC_36, KeyCode.KC_37, KeyCode.KC_38,
			KeyCode.KC_39, KeyCode.KC_40 };
	private static final KeyCode	ROW_3_ENTER	= KeyCode.KC_28;
	private static final KeyCode	ROW_4_LEFT_SHIFT= KeyCode.KC_42;
	private static final KeyCode[]	ROW_4_SEQ = {
			KeyCode.KC_44, KeyCode.KC_45, KeyCode.KC_46,
			KeyCode.KC_47, KeyCode.KC_48, KeyCode.KC_49,
			KeyCode.KC_50, KeyCode.KC_51, KeyCode.KC_52,
			KeyCode.KC_53 };
	private static final KeyCode	ROW_4_RIGHT_SHIFT= KeyCode.KC_54;
	private static final KeyCode	ROW_5_LEFT_CTRL	= KeyCode.KC_29;
	private static final KeyCode	ROW_5_LEFT_ALT	= KeyCode.KC_56;
	private static final KeyCode	ROW_5_SPACE	= KeyCode.KC_57;
	private static final KeyCode	ROW_5_RIGHT_ALT	= KeyCode.KC_100;
	private static final KeyCode	ROW_5_RIGHT_CTRL= KeyCode.KC_97;

	// keycodes for numeric keypad keys
	private static final KeyCode	NUM_NUMLOCK	= KeyCode.KC_69;
	private static final KeyCode	NUM_SLASH	= KeyCode.KC_98;
	private static final KeyCode	NUM_ASTERISK	= KeyCode.KC_55;
	private static final KeyCode	NUM_MINUS	= KeyCode.KC_74;
	private static final KeyCode	NUM_7		= KeyCode.KC_71;
	private static final KeyCode	NUM_8		= KeyCode.KC_72;
	private static final KeyCode	NUM_9		= KeyCode.KC_73;
	private static final KeyCode	NUM_4		= KeyCode.KC_75;
	private static final KeyCode	NUM_5		= KeyCode.KC_76;
	private static final KeyCode	NUM_6		= KeyCode.KC_77;
	private static final KeyCode	NUM_PLUS	= KeyCode.KC_78;
	private static final KeyCode	NUM_1		= KeyCode.KC_79;
	private static final KeyCode	NUM_2		= KeyCode.KC_80;
	private static final KeyCode	NUM_3		= KeyCode.KC_81;
	private static final KeyCode	NUM_0		= KeyCode.KC_82;
	private static final KeyCode	NUM_PERIOD	= KeyCode.KC_83;
	private static final KeyCode	NUM_ENTER	= KeyCode.KC_96;

	// standard colours for various conditions
	private static final Color	COLOUR_PANEL	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.panel_background" ));
	private static final Color	COLOUR_NEUTRAL	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.neutral" ));
	private static final Color	COLOUR_EXPECTED	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.expected" ));
	private static final Color	COLOUR_SUCCESS	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.success" ));
	private static final Color	COLOUR_ERROR	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.error" ));

	// associate some colours with fingers
	private static final Color	COLOUR_LH_PINKY	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.lh_pinky" ));
	private static final Color	COLOUR_LH_RING	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.lh_ring" ));
	private static final Color	COLOUR_LH_MIDDLE=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.lh_middle" ));
	private static final Color	COLOUR_LH_INDEX	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.lh_index" ));
	private static final Color	COLOUR_RH_INDEX	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.rh_index" ));
	private static final Color	COLOUR_RH_MIDDLE=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.rh_middle" ));
	private static final Color	COLOUR_RH_RING	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.rh_ring" ));
	private static final Color	COLOUR_RH_PINKY	=
			Colour.find_colour( colours.getString(
						"KeyboardGUI.rh_pinky" ));

	// a note of the keyboard layout we are monitoring
	private KBLayout	kblayout;//		= null;

	// tables of keyicons by keycode, and fingericons by finger
	private Hashtable	keyicons_by_keycode	= new Hashtable();
	private Hashtable	fingericons_by_finger	= new Hashtable();

	// table of colours by finger, for associations when learning
	private Hashtable	colours_by_finger	= new Hashtable();

	// additional table of keyicons for the numeric keypad
	private Hashtable	numkeycodes_by_char	= new Hashtable();

	// keyicons that are currently coloured up in some way
	private KeyIcon		last_expected;//	= null;
	private KeyIcon		num_last_expected;//	= null;
	private KeyIcon		last_error;//		= null;
	private KeyIcon		num_last_error;//	= null;
	private KeyIcon		last_success;//		= null;
	private KeyIcon		num_last_success;//	= null;

	// timestamps on the last error and success highlights
	private long		timestamp_error;//	= 0;
	private long		timestamp_success;//	= 0;

	// fingericon that is currently coloured up as being expected
	private FingerIcon	last_expected_finger;//	= null;

	// semi-constant keyicons, for shift and meta_alt keys
	private KeyIcon		left_shift;//		= null;
	private KeyIcon		right_shift;//		= null;
	private KeyIcon		left_alt;//		= null;
	private KeyIcon		right_alt;//		= null;

	// synchronization monitor, for setVisible notifications
	private Object		visible		= new Object();


	/**
	 * Initialize the whole GUI set.
	 */
	public KeyboardGUI( boolean inAnApplet, KBLayout kblayout ) {
		this.addWindowListener( this );
		this.inAnApplet = inAnApplet;

		// note our kblayout
		this.kblayout = kblayout;

		// create GUI components
		setTitle( messages.getString( "KeyboardGUI.title" ));
		setLayout( new BorderLayout() );
		setForeground( Colour.find_colour(
			colours.getString( "KeyboardGUI.foreground" )));
		setBackground( Colour.find_colour(
			colours.getString( "KeyboardGUI.background" )));
		setFont( Font.decode(
			colours.getString( "KeyboardGUI.font" )));

		// create the menu bar stuff
		menu_bar = new MenuBar();
		setMenuBar( menu_bar );
		// File menu
		file_menu = new Menu(
			messages.getString( "KeyboardGUI.file" ));
		menu_bar.add( file_menu );
		file_close = new MenuItem(
			messages.getString( "KeyboardGUI.file_close" ),
			new MenuShortcut(
				messages.getString(
				"KeyboardGUI.file_close_sc" ).charAt(0) ));
		file_close.addActionListener( this );
		file_menu.add( file_close );
		// Option menu
		options_menu = new Menu(
			messages.getString( "KeyboardGUI.options" ));
		menu_bar.add( options_menu );
		options_expected = new CheckboxMenuItem(
			messages.getString( "KeyboardGUI.options_expected" ),
				true );
		options_expected.addItemListener( this );
		options_menu.add( options_expected );
		options_colourize = new CheckboxMenuItem(
			messages.getString( "KeyboardGUI.options_colourize" ),
				true );
		options_colourize.addItemListener( this );
		options_menu.add( options_colourize );
		options_legends = new CheckboxMenuItem(
			messages.getString( "KeyboardGUI.options_legends" ),
				true );
		options_legends.addItemListener( this );
		options_menu.add( options_legends );
		options_error = new CheckboxMenuItem(
			messages.getString( "KeyboardGUI.options_error" ),
				true );
		options_error.addItemListener( this );
		options_menu.add( options_error );
		options_success = new CheckboxMenuItem(
			messages.getString( "KeyboardGUI.options_success" ),
				false );
		options_success.addItemListener( this );
		options_menu.add( options_success );

		// set up the associations between fingers and colours;
		// unspecified ones are left neutral
		colours_by_finger.put( Finger.LH_PINKY,  COLOUR_LH_PINKY );
		colours_by_finger.put( Finger.LH_RING,   COLOUR_LH_RING );
		colours_by_finger.put( Finger.LH_MIDDLE, COLOUR_LH_MIDDLE );
		colours_by_finger.put( Finger.LH_INDEX,  COLOUR_LH_INDEX );
		colours_by_finger.put( Finger.RH_INDEX,  COLOUR_RH_INDEX );
		colours_by_finger.put( Finger.RH_MIDDLE, COLOUR_RH_MIDDLE );
		colours_by_finger.put( Finger.RH_RING,   COLOUR_RH_RING );
		colours_by_finger.put( Finger.RH_PINKY,  COLOUR_RH_PINKY );

		// create the panel display that contains all the keyicons
		// and fingericons for the kblayout we have to work with,
		Panel panel = make_kb_panel();
		panel.setForeground( Colour.find_colour(
			colours.getString( "KeyboardGUI.panel_foreground" )));
		panel.setBackground( Colour.find_colour(
			colours.getString( "KeyboardGUI.panel_background" )));
		panel.setFont( Font.decode(
			colours.getString( "KeyboardGUI.panel_font" )));

		// add the graphics display to the main panel
		add( "Center", panel );

		// create the keyboard type indicator label
		String[] fmtargs = { kblayout.get_kb_id() };
		Label kbtype_label = new Label(
			MessageFormat.format(
				messages.getString( "KeyboardGUI.kbtype" ),
					fmtargs ), Label.RIGHT );
		add( "South", kbtype_label );

		// set up left and right shift and alt keyicons to save many
		// lookups later
		left_shift	= (KeyIcon)(keyicons_by_keycode.get(
					ROW_4_LEFT_SHIFT ));
		right_shift	= (KeyIcon)(keyicons_by_keycode.get(
					ROW_4_RIGHT_SHIFT ));
		left_alt	= (KeyIcon)(keyicons_by_keycode.get(
					ROW_5_LEFT_ALT ));
		right_alt	= (KeyIcon)(keyicons_by_keycode.get(
					ROW_5_RIGHT_ALT ));

		// set up the numeric keypad table, so that certain characters
		// get shown on that keypad as well; note that this is pretty
		// static - we just assume the numeric keypad doesn't vary
		// for different keyboard types
		numkeycodes_by_char.put( new Character( '/' ),  NUM_SLASH    );
		numkeycodes_by_char.put( new Character( '*' ),  NUM_ASTERISK );
		numkeycodes_by_char.put( new Character( '-' ),  NUM_MINUS    );
		numkeycodes_by_char.put( new Character( '7' ),  NUM_7        );
		numkeycodes_by_char.put( new Character( '8' ),  NUM_8        );
		numkeycodes_by_char.put( new Character( '9' ),  NUM_9        );
		numkeycodes_by_char.put( new Character( '4' ),  NUM_4        );
		numkeycodes_by_char.put( new Character( '5' ),  NUM_5        );
		numkeycodes_by_char.put( new Character( '6' ),  NUM_6        );
		numkeycodes_by_char.put( new Character( '+' ),  NUM_PLUS     );
		numkeycodes_by_char.put( new Character( '1' ),  NUM_1        );
		numkeycodes_by_char.put( new Character( '2' ),  NUM_2        );
		numkeycodes_by_char.put( new Character( '3' ),  NUM_3        );
		numkeycodes_by_char.put( new Character( '0' ),  NUM_0        );
		numkeycodes_by_char.put( new Character( '.' ),  NUM_PERIOD   );
		numkeycodes_by_char.put( new Character( '\n' ), NUM_ENTER    );

		// colourize the keyboard display, so that the associations
		// between fingers and colours can be seen if the option is set
		colourize_keyicons();
		colourize_fingericons();

		// call the method to ensure that legends are being displayed
		// according to the menu options setting
		display_legends();

		// start the timeout thread
		new Thread( this, "KeyboardGUI_timeout" ).start();
	}


	/**
	 * Thread to clear highlighting after a preset period.
	 */
	public void run() {

		// thread 'tick' period, in mS, and error/success timeouts
		final int	THREAD_TICK		= 250;
		final int	HIGHLIGHT_TIMEOUT	=
				Integer.parseInt( messages.getString(
					"KeyboardGUI.highlighttimeout" ));

		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// if our window isn't visible then just do nothing
			while ( ! isVisible() ) {
				try { synchronized( visible ) {
					visible.wait();
				}}
				catch ( InterruptedException e ) {}
			}

			// wait until the next tick
			try { Thread.sleep( THREAD_TICK ); }
			catch ( InterruptedException e ) {}

			// get the current system time
			long now = System.currentTimeMillis();

			// if the error highlight has been up long enough then
			// drop it
			if ( now - timestamp_error >= HIGHLIGHT_TIMEOUT )
				clear_error();

			// do the same for then success highlight
			if ( now - timestamp_success >= HIGHLIGHT_TIMEOUT )
				clear_success();
		}
	}


	/*
	 * Create the complete keyboard display panel.
	 */
	private Panel make_kb_panel() {

		// create the overall panel, and set six rows
		Panel panel = new Panel();
		panel.setLayout( new GridLayout( 0, 1, 0, 0 ));

		// preload some localized strings
		String bkspc	= messages.getString( "KeyboardGUI.bkspc" );
		String tab	= messages.getString( "KeyboardGUI.tab" );
		String caps	= messages.getString( "KeyboardGUI.caps" );
		String enter	= messages.getString( "KeyboardGUI.enter" );
		String shift	= messages.getString( "KeyboardGUI.shift" );
		String ctrl	= messages.getString( "KeyboardGUI.ctrl" );
		String alt	= messages.getString( "KeyboardGUI.alt" );
		String kpNm	= messages.getString( "KeyboardGUI.kpnm" );
		String kpEn	= messages.getString( "KeyboardGUI.kpen" );
		String fL	= messages.getString( "KeyboardGUI.fl" );
		String fR	= messages.getString( "KeyboardGUI.fr" );
		String fM	= messages.getString( "KeyboardGUI.fm" );
		String fI	= messages.getString( "KeyboardGUI.fi" );
		String fT	= messages.getString( "KeyboardGUI.ft" );

		// create the top row of keys
		Panel row_1 = new Panel();
		row_1.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_1.add( new_keyicon( "", 100, ROW_1_KEY1 ));
		for ( int index = 0; index < ROW_1_SEQ.length; index++ )
			row_1.add( new_keyicon( "", 100, ROW_1_SEQ[ index ] ));
		row_1.add( new_keyicon( bkspc, 200, ROW_1_BACKSPACE ));

		// create the second row of keys
		Panel row_2 = new Panel();
		row_2.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_2.add( new_keyicon( tab, 150, ROW_2_TAB ));
		for ( int index = 0; index < ROW_2_SEQ.length; index++ )
			row_2.add( new_keyicon( "", 100, ROW_2_SEQ[ index ] ));
		row_2.add( new_keyicon( "", 150, ROW_2_KEY14 ));

		// create the third row
		Panel row_3 = new Panel();
		row_3.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_3.add( new_keyicon( caps, 175, ROW_3_CAPSLOCK ));
		for ( int index = 0; index < ROW_3_SEQ.length; index++ )
			row_3.add( new_keyicon( "", 100, ROW_3_SEQ[ index ] ));
		row_3.add( new_keyicon( enter, 225, ROW_3_ENTER ));

		// create the fourth row
		Panel row_4 = new Panel();
		row_4.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_4.add( new_keyicon( shift, 225, ROW_4_LEFT_SHIFT ));
		for ( int index = 0; index < ROW_4_SEQ.length; index++ )
			row_4.add( new_keyicon( "", 100, ROW_4_SEQ[ index ] ));
		row_4.add( new_keyicon( shift, 275, ROW_4_RIGHT_SHIFT ));

		// create the bottom keyboard row
		Panel row_5 = new Panel();
		row_5.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_5.add( new_keyicon( ctrl, 150, ROW_5_LEFT_CTRL   ));
		row_5.add( new_keyicon( null, 100, KeyCode.KC_UNUSED ));//spce
		row_5.add( new_keyicon( alt,  150, ROW_5_LEFT_ALT    ));
		row_5.add( new_keyicon( "",   700, ROW_5_SPACE       ));
		row_5.add( new_keyicon( alt,  150, ROW_5_RIGHT_ALT   ));
		row_5.add( new_keyicon( null, 100, KeyCode.KC_UNUSED ));//spce
		row_5.add( new_keyicon( ctrl, 150, ROW_5_RIGHT_CTRL));

		// now make keypad additions to the above rows; at the end
		// of each row, add a spacer, then the relevant numeric keypad
		// keys
		row_1.add( new_keyicon( null,  200, KeyCode.KC_UNUSED  ));//spce
		row_1.add( new_keyicon( kpNm,  100, NUM_NUMLOCK        ));
		row_1.add( new_keyicon( "/",   100, NUM_SLASH          ));
		row_1.add( new_keyicon( "*",   100, NUM_ASTERISK       ));
		row_1.add( new_keyicon( "-",   100, NUM_MINUS          ));
		row_2.add( new_keyicon( null,  200, KeyCode.KC_UNUSED  ));//spce
		row_2.add( new_keyicon( "7",   100, NUM_7              ));
		row_2.add( new_keyicon( "8",   100, NUM_8              ));
		row_2.add( new_keyicon( "9",   100, NUM_9              ));
		row_2.add( new_keyicon( "",    100, KeyCode.KC_UNUSED  ));//spce
		row_3.add( new_keyicon( null,  200, KeyCode.KC_UNUSED  ));//spce
		row_3.add( new_keyicon( "4",   100, NUM_4              ));
		row_3.add( new_keyicon( "5",   100, NUM_5              ));
		row_3.add( new_keyicon( "6",   100, NUM_6              ));
		row_3.add( new_keyicon( "+",   100, NUM_PLUS           ));
		row_4.add( new_keyicon( null,  200, KeyCode.KC_UNUSED  ));//spce
		row_4.add( new_keyicon( "1",   100, NUM_1              ));
		row_4.add( new_keyicon( "2",   100, NUM_2              ));
		row_4.add( new_keyicon( "3",   100, NUM_3              ));
		row_4.add( new_keyicon( "",    100, KeyCode.KC_UNUSED  ));//spce
		row_5.add( new_keyicon( null,  200, KeyCode.KC_UNUSED  ));//spce
		row_5.add( new_keyicon( "0",   200, NUM_0              ));
		row_5.add( new_keyicon( ".",   100, NUM_PERIOD         ));
		row_5.add( new_keyicon( kpEn,  100, NUM_ENTER          ));

		// the finger icons row
		Panel left_finger_row	= new Panel();
		Panel right_finger_row	= new Panel();
		left_finger_row.setLayout(  new FlowLayout(
						FlowLayout.CENTER, 0, 0 ) );
		right_finger_row.setLayout( new FlowLayout(
						FlowLayout.CENTER, 0, 0 ) );
		left_finger_row.add(  new_fingericon( fL, Finger.LH_PINKY  ));
		left_finger_row.add(  new_fingericon( fR, Finger.LH_RING   ));
		left_finger_row.add(  new_fingericon( fM, Finger.LH_MIDDLE ));
		left_finger_row.add(  new_fingericon( fI, Finger.LH_INDEX  ));
		left_finger_row.add(  new_fingericon( fT, Finger.LH_THUMB  ));
		right_finger_row.add( new_fingericon( fT, Finger.RH_THUMB  ));
		right_finger_row.add( new_fingericon( fI, Finger.RH_INDEX  ));
		right_finger_row.add( new_fingericon( fM, Finger.RH_MIDDLE ));
		right_finger_row.add( new_fingericon( fR, Finger.RH_RING   ));
		right_finger_row.add( new_fingericon( fL, Finger.RH_PINKY  ));
		Panel finger_row = new Panel();
		finger_row.setLayout( new GridLayout( 0, 3, 0, 0 ));
		finger_row.add( left_finger_row );
		finger_row.add( right_finger_row );
		// add a space panel so the finger panels line up pretty much
		// underneath the main keyboard display
		finger_row.add( new Panel() );

		// add the rows to the main panel
		panel.add( row_1 ); panel.add( row_2 ); panel.add( row_3 );
		panel.add( row_4 ); panel.add( row_5 ); panel.add( finger_row );

		// return the main panel to the caller
		return ( panel );
	}

	/*
	 * Add a new keyicon with either the correct legend for the key
	 * from the keyboard layout, or the given legend, and width factor,
	 * for the given keycode.
	 */
	private KeyIcon new_keyicon( String default_legend, int width_factor,
			KeyCode keycode ) {

		// determine the legend - start by finding the glyph for
		// this keycode, preferable the unshifted one
		Glyph glyph = kblayout.find_glyph( keycode,
						KeyShiftState.UNSHIFTED );
		if ( glyph == null )
			glyph = kblayout.find_glyph( keycode,
						KeyShiftState.SHIFTED );
		if ( glyph == null )
			glyph = kblayout.find_glyph( keycode,
						KeyShiftState.META_ALT );

		// if there is no glyph, use the default legend passed in
		String legend = default_legend;
		if ( glyph != null ) {

			// for space and return, use the glyph's text,
			// otherwise just use the key character directly
			Character character = glyph.get_character();
			if ( character.charValue() == ' ' )
				legend = messages.getString(
							"KeyboardGUI.space" );
			else if ( character.charValue() == '\n' )
				legend = messages.getString(
							"KeyboardGUI.return" );
			else
				legend = character.toString();
		}

		// create the new key icon with this legend, and give it
		// an initial neutral colour
		KeyIcon keyicon = new KeyIcon( legend, width_factor );
		keyicon.set_colour( COLOUR_NEUTRAL );

		// add the new keyicon to the table for this keycode
		keyicons_by_keycode.put( keycode, keyicon );

		// add a tooltip for the keyicon, giving more details about
		// what that key actually is
		add_keyicon_tooltip( keycode, keyicon );

		// return it so it can be added to the graphics
		return( keyicon );
	}

	/*
	 * Decorate a keyicon with tooltips giving more information about it.
	 */
	private void add_keyicon_tooltip( KeyCode keycode, KeyIcon keyicon ) {

		// look up the possible glyphs for this keycode
		Glyph unshifted_glyph = kblayout.find_glyph( keycode,
						KeyShiftState.UNSHIFTED );
		Glyph shifted_glyph = kblayout.find_glyph( keycode,
						KeyShiftState.SHIFTED );
		Glyph metaalt_glyph = kblayout.find_glyph( keycode,
						KeyShiftState.META_ALT );

		// now find the three characters for these glyphs
		char unshifted_char	= ( unshifted_glyph != null ) ?
				unshifted_glyph.get_character().charValue() : 0;
		char shifted_char	= ( shifted_glyph != null ) ?
				shifted_glyph.get_character().charValue() : 0;
		char metaalt_char	= ( metaalt_glyph != null ) ?
				metaalt_glyph.get_character().charValue() : 0;

		// string these characters together into a string
		StringBuffer tooltip_info = new StringBuffer();
		if ( unshifted_char != 0 ) {
			if ( unshifted_char == ' ' )
				tooltip_info.append( messages.getString(
							"KeyboardGUI.space" ));
			else if ( unshifted_char == '\n' )
				tooltip_info.append( messages.getString(
							"KeyboardGUI.return" ));
			else {
				tooltip_info.append( "'" );
				tooltip_info.append( unshifted_char );
				tooltip_info.append( "'" );
			}
		}
		if ( shifted_char != 0 ) {
			if ( tooltip_info.length() > 0 )
				tooltip_info.append( ',' );
			tooltip_info.append( "'" );
			tooltip_info.append( shifted_char );
			tooltip_info.append( "'" );
		}
		if ( metaalt_char != 0 ) {
			if ( tooltip_info.length() > 0 )
				tooltip_info.append( ',' );
			tooltip_info.append( "'" );
			tooltip_info.append( metaalt_char );
			tooltip_info.append( "'" );
		}

		// if we have any details, and we should after all of that,
		// create a tooltip with them
		if ( tooltip_info.length() != 0 )
			new Tooltip( this, keyicon, tooltip_info.toString() );
	}

	/*
	 * Add a new finger with the given legend and finger.
	 */
	private FingerIcon new_fingericon( String legend, Finger finger ) {

		// create the new finger icon
		FingerIcon fingericon = new FingerIcon( legend );
		fingericon.set_colour( COLOUR_NEUTRAL );

		// add it to the table for this keycode
		fingericons_by_finger.put( finger, fingericon );

		// add a tooltip for the fingericon, giving more details about
		add_fingericon_tooltip( finger, fingericon );

		// return it so it can be added to the graphics
		return( fingericon );
	}

	/*
	 * Decorate a fingericon with tooltips giving more information about it.
	 */
	private void add_fingericon_tooltip( Finger finger,
						FingerIcon fingericon ) {

		// find the string name for this finger
		String finger_name = messages.getString(
						finger.get_resource_id() );

		// add the tooltip with this information
		new Tooltip( this, fingericon, finger_name );
	}


	/**
	 * Colourize the background of keyicons to show the associations
	 * between fingers and keys, depending on the option setting.
	 */
	private void colourize_keyicons() {

		// note the state of the colourizing option
		boolean do_colours = options_colourize.getState();

		// iterate over the keymap, setting colours for everything
		// that is represented; in practice, these are all we can
		// say for sure we know what the fingering of these keys is
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)(enum.nextElement());

			// discover the keycode and finger for this glyph
			KeyCode keycode		= glyph.get_keycode();
			Finger finger		= glyph.get_finger();

			// find the keyicon that represents this keycode
			KeyIcon keyicon = (KeyIcon)(keyicons_by_keycode.get(
								keycode ));
			if ( keyicon != null ) {

				// found a keyicon; should we do colours?
				if ( do_colours ) {

					// colour-code the keyicon for the
					// finger, if it has an association
					Color colour = (Color)
						(colours_by_finger.get( finger ));
					if ( colour != null )
						keyicon.set_bgcolour( colour );
				}
				else
					// no colours required, so set the
					// background to match the panel
					keyicon.set_bgcolour( COLOUR_PANEL );
			}
		}
	}

	/**
	 * Colourize the background of fingericons to show the associations
	 * between fingers and keys, depending on the option setting.
	 */
	private void colourize_fingericons() {

		// note the state of the colourizing option
		boolean do_colours = options_colourize.getState();

		// iterate over every finger known
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)(enum.nextElement());

			// discover the fingericon that does this finger
			FingerIcon fingericon = (FingerIcon)
					(fingericons_by_finger.get( finger ));
			if ( fingericon != null ) {

				// fingericon found; should we do colours?
				if ( do_colours ) {

					// colour-code the keyicon for the
					// finger, if it has an association
					Color colour = (Color)
						(colours_by_finger.get( finger ));
					if ( colour != null )
						fingericon.set_colour( colour );
				}
				else
					// no colours required, so set the
					// foreground to a neutral colour
					fingericon.set_colour( COLOUR_NEUTRAL );
			}
		}
	}


	/**
	 * Display or hide legends, depending on the option setting.
	 */
	private void display_legends() {

		// note the state of the legends option
		boolean do_legends = options_legends.getState();

		// iterate over every keyicon and fingericon, and for each
		// one found, set its show_legend flag to match the value
		// of our GUI's menu option
		for ( Enumeration enum = keyicons_by_keycode.elements();
				enum.hasMoreElements(); ) {
			KeyIcon keyicon = (KeyIcon)(enum.nextElement());

			// set the keyicon legend flag to match the option
			keyicon.show_legend( do_legends );
		}
		for ( Enumeration enum = fingericons_by_finger.elements();
				enum.hasMoreElements(); ) {
			FingerIcon fingericon =
					 (FingerIcon)(enum.nextElement());

			// set the keyicon legend flag to match the option
			fingericon.show_legend( do_legends );
		}
	}

	/*
	 * Set the keyicon for a given keycode, glyph, or character
	 * to a given colour.
	 */
	private KeyIcon highlight_keyicon( KeyCode keycode, Color colour ) {

		// look up the keyicon for this keycode
		KeyIcon keyicon = (KeyIcon)(keyicons_by_keycode.get( keycode ));
		if ( keyicon != null ) {

			// found a keyicon; highlight it to the requested colour
			keyicon.set_highlight( colour );

			// return the keyicon just highlighted
			return ( keyicon );
		}

		// no keyicon found for this keycode, so return null
		return ( null );
	}
	private KeyIcon highlight_keyicon( Glyph glyph, Color colour ) {

		// send back the result of simply highlighting the keyicon
		// for the given glyph's keycode
		return ( highlight_keyicon( glyph.get_keycode(), colour ));
	}
	private KeyIcon highlight_keyicon( char character, Color colour ) {

		// look up the glyph for this character
		Glyph glyph = kblayout.find_glyph( character );
		if ( glyph != null )

			// found a glyph for this character, so return the
			// result of highlighting this glyph
			return ( highlight_keyicon( glyph, colour ));

		// no match found for the given glyph
		return ( null );
	}
	private KeyIcon num_highlight_keyicon( char character, Color colour ) {

		// see if this key could also be on the numeric keypad
		KeyCode num_keycode = (KeyCode)(numkeycodes_by_char.get(
						new Character( character )));
		if ( num_keycode != null )

			// found a numeric keycode for this character, so
			// send back the result of highlighting that keyicon
			return ( highlight_keyicon( num_keycode, colour ));

		// there is no numeric keycode that matches this character
		return ( null );
	}


	/**
	 * Indicate a keyicon and fingericon as the next key expected.
	 */
	public void char_expected( char character ) {

		// ignore the call if the option is not set
		if ( options_expected.getState() ) {

			// clear any previous expectations
			clear_expected();

			// get the glyph for this character
			Glyph glyph = kblayout.find_glyph( character );
			if ( glyph != null ) {

				// identify the fingericon for this glyph
				Finger finger		= glyph.get_finger();
				FingerIcon fingericon	= (FingerIcon)
					(fingericons_by_finger.get( finger ));
				if ( fingericon != null )

					// fingericon found, so highlight it
					fingericon.set_highlight(
							COLOUR_EXPECTED );
				last_expected_finger	= fingericon;

				// if a shift is required with this char then
				// highlight that key too
				KeyIcon shifticon = null;
				KeyShiftState keyshiftstate =
						glyph.get_shiftstate();
				if ( keyshiftstate == KeyShiftState.SHIFTED )
					if ( finger.is_left_handed() )
						shifticon = right_shift;
					else
						shifticon = left_shift;
				else
				if ( keyshiftstate == KeyShiftState.META_ALT )
					if ( finger.is_left_handed() )
						shifticon = right_alt;
					else
						shifticon = left_alt;

				// highlight any shift/alt keyicon found
				if ( shifticon != null )
					shifticon.set_highlight(
							COLOUR_EXPECTED );

				// set a main keyboard keyicon to a new colour
				last_expected = highlight_keyicon( glyph,
							COLOUR_EXPECTED );
			}

			// independent of all of that, try setting a keyicon on
			// the numeric keypad
			num_last_expected = num_highlight_keyicon( character,
							COLOUR_EXPECTED );
		}
	}

	/**
	 * Clear keyicon and fingericon expectations.
	 */
	public void clear_expected() {

		// reset the last keyicons to background
		if ( last_expected != null ) {
			last_expected.clear_highlight();
			last_expected = null;
		}
		if ( num_last_expected != null ) {
			num_last_expected.clear_highlight();
			num_last_expected = null;
		}

		// clear both shift and alt keys
		left_shift.clear_highlight();	right_shift.clear_highlight();
		left_alt.clear_highlight();	right_alt.clear_highlight();

		// clear the last finger highlighting
		if ( last_expected_finger != null ) {
			last_expected_finger.clear_highlight();
			last_expected_finger = null;
		}
	}

	/**
	 * Indicate a keyicon as pressed in error.  This highlighting will
	 * automatically timeout.
	 */
	public void char_error( char character ) {

		// ignore the call if the option is not set
		if ( options_error.getState() ) {

			// clear any previous errors
			clear_error();

			// set the right keyicon(s) to a new colour
			last_error	= highlight_keyicon( character,
								COLOUR_ERROR );
			num_last_error	= num_highlight_keyicon( character,
								COLOUR_ERROR );

			// note the time, for timeouts
			timestamp_error = System.currentTimeMillis();
		}
	}

	/**
	 * Clear keyicon error.  If the error keyicon is also the current
	 * expected keyicon, this will not change the keyicon colour.
	 */
	public void clear_error() {

		// reset the last keyicons to background
		if ( last_error != null ) {
			if ( last_error != last_expected )
				last_error.clear_highlight();
			last_error = null;
		}
		if ( num_last_error != null ) {
			if ( num_last_error != num_last_expected )
				num_last_error.clear_highlight();
			num_last_error = null;
		}
	}

	/**
	 * Indicate a keyicon as pressed correctly.  This highlighting will
	 * automatically timeout.
	 */
	public void char_success( char character ) {

		// ignore the call if the option is not set
		if ( options_success.getState() ) {

			// clear any previous success
			clear_success();

			// set the right keyicon(s) to a new colour
			last_success	= highlight_keyicon( character,
							COLOUR_SUCCESS );
			num_last_success= num_highlight_keyicon( character,
							COLOUR_SUCCESS );

			// if this is the current expected keyicon, forget
			// the expectation so the success highlight isn't lost;
			// this is a normal case - the expected key was pressed,
			// and the expectation moving on will otherwise erase
			// our highlighting
			if ( last_success == last_expected )
				last_expected = null;
			if ( num_last_success == num_last_expected )
				num_last_expected = null;

			// note the time, for timeouts
			timestamp_success = System.currentTimeMillis();
		}
	}

	/**
	 * Clear keyicon success.  If the success keyicon is also the current
	 * expected keyicon, this will not change the keyicon colour.
	 */
	public void clear_success() {

		// reset the last keyicons to background
		if ( last_success != null ) {
			if ( last_success != last_expected )
				last_success.clear_highlight();
			last_success = null;
		}
		if ( num_last_success != null ) {
			if ( num_last_success != num_last_expected )
				num_last_success.clear_highlight();
			num_last_success = null;
		}
	}

	/**
	 * Override setVisible to ensure we restart our thread when visible.
	 */
	public void setVisible( boolean is_visible ) {
		super.setVisible( is_visible );
		if ( is_visible )
			synchronized( visible ) { visible.notify(); }
	}

	/*
	 * Handle dismiss, close, etc by just hiding ourselves.
	 */
	private void handle_close_button() {

		// simply just hide this frame from view
		setVisible( false );
	}

	/*
	 * Handle GUI component actions and events.
	 */
	public void actionPerformed( ActionEvent evt ) {
		if ( evt.getSource() instanceof MenuItem ) {
			if ( evt.getSource() == file_close )
				handle_close_button();
		}
	}
	public void windowClosing( WindowEvent evt ) {
		handle_close_button();
	}
	public void windowClosed( WindowEvent evt ) {}
	public void windowOpened( WindowEvent evt ) {}
	public void windowIconified( WindowEvent evt ) {}
	public void windowDeiconified( WindowEvent evt ) {}
	public void windowActivated( WindowEvent evt ) {}
	public void windowDeactivated( WindowEvent evt ) {}
	public void itemStateChanged( ItemEvent evt ) {
		if ( evt.getSource() instanceof CheckboxMenuItem ) {
			if ( evt.getSource() == options_expected )
				clear_expected();
			else if ( evt.getSource() == options_colourize ) {
				colourize_keyicons();
				colourize_fingericons();
			}
			else if ( evt.getSource() == options_legends )
				display_legends();
		}
	}

	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		KeyboardGUI kb = new KeyboardGUI( false,
				KBLayout.create_kblayout( "US" ) );
		kb.pack(); kb.setVisible( true );
		while ( true ) {
			kb.char_expected( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 2000 ); }
			catch( InterruptedException e ) {}
			kb.char_error( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 300 ); }
			catch( InterruptedException e ) {}
			kb.char_success( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 300 ); }
			catch( InterruptedException e ) {}
		}
	}
}
